//	<><><><><><><><><><><><><>  Plot3DF.cpp  <><><><><><><><><><><><><><> */
//		
// ----------------------------------------------------
// (C) Copyright Clark Dailey, Inc. 2006
//
// Implementation of Plot3DF C++ class
//
// --------------------------------------------------------------------------
//

#include "stdafx.h"	// precompiled header
#include <stdio.h>
#include "Plot3DF.h"
#include "SurfX3D.h"

// ------------------------------------------------------------
Plot3DF::Plot3DF()
{
	// use default values
	DefInputs();

	// initialize dynamic memory
	m_pMask1  = 0;
	m_pMask2  = 0;
	m_mask_size = 0;
	m_nXps    = 0;
	m_nYps    = 0;
	m_nZps    = 0;

	m_pFuncPlotCalc = new PlotCalc;
	m_pRPlotCalc    = new PlotCalc;
	m_pGPlotCalc    = new PlotCalc;
	m_pBPlotCalc    = new PlotCalc;
	m_use_sphere_range=0;
	m_sphere_radius=10;
	SetIniHeader("Implicit");
}

// ------------------------------------------------------------
Plot3DF::~Plot3DF()
{
	DeleteMaskArrays();

	delete m_pFuncPlotCalc;
	delete m_pRPlotCalc;
	delete m_pGPlotCalc;
	delete m_pBPlotCalc;
	m_pFuncPlotCalc = 0;
	m_pRPlotCalc    = 0;
	m_pGPlotCalc    = 0;
	m_pBPlotCalc    = 0;
}

// ------------------------------------------------------------
void Plot3DF::DefInputs(void)
{
	// initialize surface equation
	m_sSurfaceName   = "Untitled";
 	m_sFuncEquation  = "(x^2+y^2+z^2-c1-c2)^2+(4*c1*(x^2-c2))";
 	m_sREquation     = "(sin(2*x)+1)*50";
 	m_sGEquation     = "(cos(2*x)+1)*50";
 	m_sBEquation     = "0";
	m_sC1 = "100";
	m_sC2 = "25";
	m_sC3 = "1";
	m_sC4 = "1";
	m_sC5 = "1";
	m_sC6 = "1";
	m_sC7 = "1";
	m_sC8 = "1";

	// init parameters
	m_nzplanes=50;
	m_remove_hidden=1;
	m_draw_lines=1;
	m_clip_mode = CLIP_NONE;
	m_is_clip_inverted = 0;
}

// ------------------------------------------------------------
void Plot3DF::Reset(void)
{
	InitBase();
	DeleteMaskArrays();
	DefInputs();
}

// ------------------------------------------------------------
// returns: 0=ok
int  Plot3DF::InitEquations(void)
{
	int fcol, rcol, gcol, bcol;

	// check input parameters
	if (m_xmin >= m_xmax)
	{
		m_is_equation_ok = 0;
		theApp.SetStatusString("Error: Xmin >= Xmax");
		return(1);
	}
	if (m_ymin >= m_ymax)
	{
		m_is_equation_ok = 0;
		theApp.SetStatusString("Error: Ymin >= Ymax");
		return(1);
	}
	if (m_zmin >= m_zmax)
	{
		m_is_equation_ok = 0;
		theApp.SetStatusString("Error: Zmin >= Zmax");
		return(1);
	}

	// update calculator - load constants and program
 	fcol = m_pFuncPlotCalc->LoadProgram("F(x,y,z)",(LPCTSTR)m_sFuncEquation,1,
			(LPCTSTR)m_sC1,(LPCTSTR)m_sC2,(LPCTSTR)m_sC3,(LPCTSTR)m_sC4,
			(LPCTSTR)m_sC5,(LPCTSTR)m_sC6,(LPCTSTR)m_sC7,(LPCTSTR)m_sC8);

 	rcol = m_pRPlotCalc->LoadProgram("R(x,y,z)",(LPCTSTR)m_sREquation,1,
			(LPCTSTR)m_sC1,(LPCTSTR)m_sC2,(LPCTSTR)m_sC3,(LPCTSTR)m_sC4,
			(LPCTSTR)m_sC5,(LPCTSTR)m_sC6,(LPCTSTR)m_sC7,(LPCTSTR)m_sC8);

 	gcol = m_pGPlotCalc->LoadProgram("G(x,y,z)",(LPCTSTR)m_sGEquation,1,
			(LPCTSTR)m_sC1,(LPCTSTR)m_sC2,(LPCTSTR)m_sC3,(LPCTSTR)m_sC4,
			(LPCTSTR)m_sC5,(LPCTSTR)m_sC6,(LPCTSTR)m_sC7,(LPCTSTR)m_sC8);

 	bcol = m_pBPlotCalc->LoadProgram("B(x,y,z)",(LPCTSTR)m_sBEquation,1,
			(LPCTSTR)m_sC1,(LPCTSTR)m_sC2,(LPCTSTR)m_sC3,(LPCTSTR)m_sC4,
			(LPCTSTR)m_sC5,(LPCTSTR)m_sC6,(LPCTSTR)m_sC7,(LPCTSTR)m_sC8);

	m_is_equation_ok = (fcol | rcol | gcol | bcol) ? 0 : 1;

	// show status string
	if (fcol)
	 	m_sEquationStatus = m_pFuncPlotCalc->GetErrString();
	else if (rcol)
	 	m_sEquationStatus = m_pRPlotCalc->GetErrString();
	else if (gcol)
	 	m_sEquationStatus = m_pGPlotCalc->GetErrString();
	else
	 	m_sEquationStatus = m_pBPlotCalc->GetErrString();

	if (!m_is_equation_ok) theApp.SetStatusString(m_sEquationStatus);
	return(m_is_equation_ok?0:1);
}

// ------------------------------------------------------------
long  Plot3DF::Color(double x,double y,double z)
{
	double dred, dgreen, dblue;
	dred   = ValidateColorValue( m_pRPlotCalc->Calc(x,y,z) );
	dgreen = ValidateColorValue( m_pGPlotCalc->Calc(x,y,z) );
	dblue  = ValidateColorValue( m_pBPlotCalc->Calc(x,y,z) );
	return(RGB(PCOLOR_TO_WINCOLOR(dred), PCOLOR_TO_WINCOLOR(dgreen), PCOLOR_TO_WINCOLOR(dblue)));
}

// ------------------------------------------------------------
// returns 0=ok
int  Plot3DF::InitPlot(void)
{
	if (InitEquations()) return(1);
	PlotOutputOn();
	CalcTM();	// calculate transformation matrix
	CalcCorners();	// calculate corners of x,y plate
	CalcDeltas();
	AllocateMaskArrays();
	return(0);
}

// ------------------------------------------------------------
int  Plot3DF::AllocateMaskArrays(void)
{
	int rc=0;

	// delete any current mask arrays
	DeleteMaskArrays();

	// allocate mask arrays
	m_pMask1  = new MASK_POINT[m_mask_size];
	m_pMask2  = new MASK_POINT[m_mask_size];
	if (m_pMask1 == 0 || m_pMask2 == 0)
	{
		DeleteMaskArrays();
		rc = 1;
	}
	return(rc);
}

// ------------------------------------------------------------
void Plot3DF::DeleteMaskArrays(void)
{
	delete [] m_pMask1;
	delete [] m_pMask2;
	m_pMask1  = 0;
	m_pMask2  = 0;
}

// ------------------------------------------------------------
void Plot3DF::CalcDeltas()
{
	double   dxp, dyp, dxv, dyv;
	unsigned uxv, uyv;

	PLOT_ASSERT(m_xvmax > m_xvmin);
	PLOT_ASSERT(m_yvmax > m_yvmin);

	// range check nplanes
	if (m_nzplanes < 2) m_nzplanes = 2;

	// video ranges
	uxv = m_xvmax - m_xvmin;
	uyv = m_yvmax - m_yvmin;
	if (uxv < 2) uxv = 2;
	if (uyv < 2) uyv = 2;
	dxv = uxv;
	dyv = uyv;

	// calc number of increments
	m_nXps = uxv;
	m_nYps = uyv;
	m_nZps = m_nzplanes;
	m_mask_size = m_nXps;

	// use base values for default
	m_xpmin = m_xpbmin;
	m_xpmax = m_xpbmax;
	m_ypmin = m_ypbmin;
	m_ypmax = m_ypbmax;
	m_zpmin = m_zpbmin;
	m_zpmax = m_zpbmax;

	// calculate for spherical range
	if (m_use_sphere_range)
	{
		m_xpbmid = xpf(m_sphere_xcenter,m_sphere_ycenter,m_sphere_zcenter);
		m_ypbmid = ypf(m_sphere_xcenter,m_sphere_ycenter,m_sphere_zcenter);
		m_zpbmid = zpf(m_sphere_xcenter,m_sphere_ycenter,m_sphere_zcenter);
		m_xpmax = m_xpbmid + m_sphere_radius;
		m_xpmin = m_xpbmid - m_sphere_radius;
		m_ypmax = m_ypbmid + m_sphere_radius;
		m_ypmin = m_ypbmid - m_sphere_radius;
		m_zpmax = m_zpbmid + m_sphere_radius;
		m_zpmin = m_zpbmid - m_sphere_radius;
	}

	// ratio so image fits in the video window
	// keep 1:1 aspect ratio with x,y video axis
	dxp = m_xpmax-m_xpmin;
	dyp = m_ypmax-m_ypmin;
	if (dxp*dyv > dyp*dxv)
	{
		dyp = (dyp*dyv)/(dxv*2.);
		m_ypmin = m_ypbmid - dyp;
		m_ypmax = m_ypbmid + dyp;
	}
	else
	{
		dxp = (dxp*dxv)/(dyv*2.);
		m_xpmin = m_xpbmid - dxp;
		m_xpmax = m_xpbmid + dxp;
	}

	// apply scale factor
	if (m_scale<1) m_scale = 1;	// avoid division by zero
	dxp = (m_xpmax-m_xpmin)/(m_scale*2./100.);
	dyp = (m_ypmax-m_ypmin)/(m_scale*2./100.);
	m_xpmax = m_xpbmid + dxp;
	m_xpmin = m_xpbmid - dxp;
	m_ypmax = m_ypbmid + dyp;
	m_ypmin = m_ypbmid - dyp;

	// calc plot to video scaling
	m_p2v = ((m_scale/100.) * (double)(m_xvmax - m_xvmin)) / (double)(m_xpmax - m_xpmin);

	// guarantee zxmax > zpmin
	if (m_zpmin>=m_zpmax)
		m_zpmin = m_zpmax - 1.e-6;

	// calc increments deltas
	m_dxp = (m_xpmax-m_xpmin)/((m_scale/100.)*double(m_nXps-1));
	m_dyp = m_dxp;	// maintain proper aspect ratio
	m_dzp = (m_zpmax-m_zpmin)/((m_scale/100.)*double(m_nZps-1));

	// plotting must start from outmost of the screen inward for proper hidden line removal
	if (m_zpcorner[1]>=m_zpcorner[0])
	{
		m_xstart[0] =  m_xmax;
		m_xstart[1] =  m_xmin;
	}
	else
	{
		m_xstart[0] =  m_xmin;
		m_xstart[1] =  m_xmax;
	}
	if (m_zpcorner[2]>m_zpcorner[0])
	{
		m_ystart[0] =  m_ymax;
		m_ystart[1] =  m_ymin;
	}
	else
	{
		m_ystart[0] =  m_ymin;
		m_ystart[1] =  m_ymax;
	}
	if (m_ypcorner[0]<m_ypcorner[4])
	{
		m_zstart[0] = m_zmin;
		m_zstart[1] = m_zmax;
	}
	else
	{
		m_zstart[0] = m_zmax;
		m_zstart[1] = m_zmin;
	}
}

// --------------------
//   POINT3  POINT4
//
//   POINT1  POINT2
// --------------------

#define POINT0	 0,  0,  0	// no movement indicator
#define POINT1	 1,  0,  0
#define POINT2	 2,  1,  0
#define POINT3	 3,  0,  1
#define POINT4	 4,  1,  1

// Draw Look-up-table
typedef struct DrawLUT_s
{
	int	from;	 // 0=none 1,2,3,4
	int xv_from; // 0 or 1
	int yv_from; // 0 or 1
	int	to;		 // 0,1,2,3,4
	int xv_to;	 // 0 or 1
	int yv_to;	 // 0 or 1
} DrawLUT;

static DrawLUT drawLUT[16] =
{
   { POINT0, POINT0 },	 //  0
   { POINT1, POINT4 },	 //  1
   { POINT2, POINT3 },	 //  2
   { POINT1, POINT2 },	 //  3
   { POINT2, POINT3 },	 //  4
   { POINT2, POINT4 },	 //  5
   { POINT1, POINT4 },	 //  6
   { POINT1, POINT4 },	 //  7
   { POINT1, POINT4 },	 //  8
   { POINT2, POINT3 },	 //  9
   { POINT1, POINT3 },	 // 10
   { POINT2, POINT3 },	 // 11
   { POINT3, POINT4 },	 // 12
   { POINT2, POINT3 },	 // 13
   { POINT1, POINT4 },	 // 14
   { POINT0, POINT0 }	 // 15
};

// ------------------------------------------------------------
void Plot3DF::PlotMask(unsigned iyp)
{
	unsigned ixp,mask;
	int xv,yv,xvm1,yvm1,w1,w2;
	MASK_POINT* pMaskPlot[2][2];
	MASK_POINT* pFrom;
	MASK_POINT* pTo;

	pMaskPlot[0][0] = m_pMask1  ;
	pMaskPlot[0][1] = m_pMask1+1;
	pMaskPlot[1][0] = m_pMask2  ;
	pMaskPlot[1][1] = m_pMask2+1;
	xvm1 = m_xvmin - 1;
	yvm1 = m_yvmin - 1;
	for (ixp=1;ixp<m_nXps;ixp++)
	{
		mask = (pMaskPlot[1][0]->sf & 8) | 
			   (pMaskPlot[1][1]->sf & 4) |
			   (pMaskPlot[0][0]->sf & 2) |
			   (pMaskPlot[0][1]->sf & 1) ;

		if (drawLUT[mask].from!=0)
		{
			pFrom = pMaskPlot[drawLUT[mask].xv_from][drawLUT[mask].yv_from];
			pTo   = pMaskPlot[drawLUT[mask].xv_to  ][drawLUT[mask].yv_to  ];

			w1 = Within(pFrom->x,pFrom->y,pFrom->z);
			w2 = Within(  pTo->x,  pTo->y,  pTo->z);

			if (w1 && w2 &&	// both points must be within drawing area
			    IsVisible(pFrom->xp,pFrom->yp,pFrom->zp) &&	// both points must be visible
				IsVisible(pFrom->xp,pFrom->yp,pFrom->zp) )
			{
				xv = drawLUT[mask].xv_from + ixp + xvm1;
				yv = drawLUT[mask].yv_from + iyp + yvm1;
				FilterMoveTo(xv,yv);
				xv = drawLUT[mask].xv_to   + ixp + xvm1;
				yv = drawLUT[mask].yv_to   + iyp + yvm1;
				FilterDrawTo(xv,yv);
			}
		}

		// advance to next check point
		pMaskPlot[1][0] = pMaskPlot[1][1]++;
		pMaskPlot[0][0] = pMaskPlot[0][1]++;
	} // for ixp
}

// ------------------------------------------------------------
// overridable clipping function; return: 0=clipped, 1=point is within drawable region
int Plot3DF::Within(double x,double y,double z)
{
	int	v;
	if (m_clip_mode == CLIP_NONE) return(1);
	if (m_clip_mode == CLIP_CUBE)
	{
		v = (x>=m_xmin && x<=m_xmax && y>=m_ymin && y<=m_ymax && z>=m_zmin && z<=m_zmax) ? 1 : 0;
		if (m_is_clip_inverted) v = !v;
		return(v);
	}
	if (m_clip_mode == CLIP_SPHERE)
	{
		double dx,dy,dz,r;
		dx = x - m_sphere_xcenter;
		dy = y - m_sphere_ycenter;
		dz = z - m_sphere_zcenter;
		r = sqrt(dx*dx+dy*dy+dz*dz);
		v = (r <= m_sphere_radius) ? 1 : 0;
		if (m_is_clip_inverted) v = !v;
		return(v);
	}
	// Note: derived class handles user defined clipping
	return(1);
}

// ------------------------------------------------------------
// returns: 0=no, 1=point is visible
int Plot3DF::IsVisible(double xp,double yp,double zp)
{
	double x,y,z,f1,f2,zpsave;
	int w1,w2;

	if (m_remove_hidden==0) return(1);  // visible if no hidden line removal

	// zearch for a zero crossing outward from screen
	zpsave = zp;
	zp = zpsave - m_dzp;
	x  = xof(xp,yp,zp);
	y  = yof(xp,yp,zp);
	z  = zof(xp,yp,zp);
	f1 = Func(x,y,z);
	w1 = Within(x,y,z);
	for (zp-=m_dzp; zp>=m_zpmin; zp-=m_dzp)
	{
		x  = xof(xp,yp,zp);
		y  = yof(xp,yp,zp);
		z  = zof(xp,yp,zp);
		f2 = Func(x,y,z);
		w2 = Within(x,y,z);

		if (f1*f2<=0. && w1 && w2)
		{
			return(0);
		}
		else
		{
			f1 = f2;
			w1 = w2;
		}
	}
	return(1);
}

// ------------------------------------------------------------
// plot cross sections
int Plot3DF::Plot(CWnd* pWnd)
{
	time_t begTime, endTime;

	SetWnd(pWnd);

	theApp.SetStatusString("Render Begin");
	time(&begTime);

	if (InitPlot()) return(2);

	WinBegPlot();

	PlotColorPixels();
	DrawLines();

	DrawAxis();

	WinEndPlot();
	time(&endTime);

	// calculate render time
	unsigned minutes, seconds;
	CalcDiffTime(begTime,endTime,&minutes,&seconds);

	// show rendering time
	CString sRenderTime;
	sRenderTime.Format("Render Complete    %u:%02u (m:s)", minutes, seconds);
	theApp.SetStatusString(sRenderTime);

	return(0);
}

// ------------------------------------------------------------
int Plot3DF::DrawLines()
{
	CString sPercent;
	MASK_POINT* pMask;
	unsigned ix,iy,iz, percent_complete;
	double   xp,yp,zp;
	double   xp0,yp0;

	if (!m_draw_lines) return(0);

	// Note: incrementing xp,yp,zp is faster than table lookups
	xp0 = xv2xp(0);
	yp0 = yv2yp(0);
	for (iz=0,zp=m_zpmin; iz<m_nZps; iz++,zp+=m_dzp)
	{
		if (IsAborted()) break;
		pMask = m_pMask2;
		for (iy=0,yp=yp0; iy<m_nYps; iy++,yp+=m_dyp)
		{
			for (ix=0,xp=xp0; ix<m_nXps; ix++,xp+=m_dxp)
			{
				pMask->xp = xp;
				pMask->yp = yp;
				pMask->zp = zp;
				pMask->x  = xof(xp,yp,zp);
				pMask->y  = yof(xp,yp,zp);
				pMask->z  = zof(xp,yp,zp);
				pMask->f  = Func( pMask->x, pMask->y, pMask->z );
				pMask->sf = (pMask->f >= 0.) ? 0xffff : 0;
				pMask++;
			} // for ix
			if (iy != 0)
				PlotMask(iy);
			// swap mask arrays
			pMask    = m_pMask1;
			m_pMask1 = m_pMask2;
			m_pMask2 = pMask;
			// Note: pMask is now set for next iteration
		} // for iy

		// show percent complete status
		percent_complete = (int)(100.*(iz+1)/m_nZps);
		sPercent.Format("Draw Sections: %u%% complete", percent_complete);
		theApp.SetStatusString(sPercent);

	} // for iz
	return(0);
}

// ------------------------------------------------------------
int Plot3DF::PlotColorPixels()
{
	CString sPercent;
	long color;
	int xv,yv, percent_complete;

	if (!m_is_color) return(0);
	for (yv=m_yvmin; yv<m_yvmax; yv++)
	{
		if (IsAborted()) break;
		for (xv=m_xvmin; xv<m_xvmax; xv++)
		{
			color = CalcPixelColor(xv,yv);
			WinSetPixelColor(xv,yv,color);
		} // for xv

		// show percent complete status
		percent_complete = (int)(100.*(yv+1)/m_yvmax);
		sPercent.Format("Colorizing: %u%% complete", percent_complete);
		theApp.SetStatusString(sPercent);
	} // for yv
	return(0);
}

// ------------------------------------------------------------
long Plot3DF::CalcPixelColor(int xv,int yv)
{
	double xp,yp,zp;
	double x,y,z,f1,f2;
	int w1,w2;

	// convert to plot space
	xp = xv2xp(xv);
	yp = yv2yp(yv);
	zp = m_zpmin;

	// zearch for a zero crossing in towards screen
	x  = xof(xp,yp,zp);
	y  = yof(xp,yp,zp);
	z  = zof(xp,yp,zp);
	f1 = Func(x,y,z);
	w1 = Within(x,y,z);
	for (zp+=m_dzp; zp<=m_zpmax; zp+=m_dzp)
	{
		x  = xof(xp,yp,zp);
		y  = yof(xp,yp,zp);
		z  = zof(xp,yp,zp);
		f2 = Func(x,y,z);
		w2 = Within(x,y,z);
	
		if (f1*f2<=0. && w1 && w2)
		{
			return(Color(x,y,z));
		}
		else
		{
			f1 = f2;
			w1 = w2;
		}
	}
	return(m_back_color);
}

// ------------------------------------------------------------
// returns: 0=ok, 1=can't open file
int  Plot3DF::ReadFile()
{
	if (!DoesFileExist(m_FilePath)) return(1);

	// surface definition
	m_sSurfaceName  = GetIniStr   ("Name",          	"Untitled"    );
	m_sC1           = GetIniStr   ("C1",	          	""    );
	m_sC2           = GetIniStr   ("C2",	          	""    );
	m_sC3           = GetIniStr   ("C3",	          	""    );
	m_sC4           = GetIniStr   ("C4",	          	""    );
	m_sC5           = GetIniStr   ("C5",	          	""    );
	m_sC6           = GetIniStr   ("C6",	          	""    );
	m_sC7           = GetIniStr   ("C7",	          	""    );
	m_sC8           = GetIniStr   ("C8",	          	""    );
	m_sFuncEquation = GetIniStr   ("PlotEquation", 	    "(x^2+y^2+z^2-c1-c2)^2+(4*c1*(x^2-c2))"    );
	m_sREquation    = GetIniStr   ("RedEquation", 	    "abs(x)*20"    );
	m_sGEquation    = GetIniStr   ("GreenEquation", 	"abs(y)*20"    );
	m_sBEquation    = GetIniStr   ("BlueEquation", 	    "0"    );

	// common parameters
	m_zrotate       = GetIniDouble("ZRotate",       	30.0  );
	m_xytilt        = GetIniDouble("XYTilt",        	60.0  );
	m_scale         = GetIniDouble("PlotScale",        100.0  );
	m_line_color    = GetIniLong  ("LineColor",   DEF_LINE_COLOR  );
	m_back_color    = GetIniLong  ("BackColor",   DEF_BKGND_COLOR );
	m_is_color      = GetIniLong  ("ShowColor",     	 1    );
	m_remove_hidden = GetIniLong  ("RemoveHidden",  	 1    );
	m_show_axis     = GetIniLong  ("ShowAxis",      	 0    );

	// specific parameters
	m_xmin          = GetIniDouble("XMin",              -10.  );
	m_xmax          = GetIniDouble("XMax",               10.  );
	m_ymin          = GetIniDouble("YMin",              -10.  );
 	m_ymax          = GetIniDouble("YMax",               10.  );
	m_zmin          = GetIniDouble("ZMin",              -10.  );
 	m_zmax          = GetIniDouble("ZMax",               10.  );
 	m_draw_lines      = GetIniLong   ("DrawLines",        1   );
	m_nzplanes        = GetIniLong   ("nZPlanes",        30   );
	m_use_sphere_range= GetIniLong   ("UseSpherical",     0   );
	m_sphere_xcenter  = GetIniDouble ("SphereXCenter",    0.  );
	m_sphere_ycenter  = GetIniDouble ("SphereYCenter",    0.  );
	m_sphere_zcenter  = GetIniDouble ("SphereZCenter",    0.  );
	m_sphere_radius   = GetIniDouble ("SphereRadius",    10.  );
	m_clip_mode       = GetIniLong   ("ClipMode",         0   );
	m_is_clip_inverted= GetIniLong   ("InvertClip",       0   );

	return(0);
}

// ------------------------------------------------------------
int  Plot3DF::WriteFile()
{
	// surface definition
	SetIniStr   ("Name",          m_sSurfaceName);
	SetIniStr   ("C1",	          m_sC1);
	SetIniStr   ("C2",	          m_sC2);
	SetIniStr   ("C3",	          m_sC3);
	SetIniStr   ("C4",	          m_sC4);
	SetIniStr   ("C5",	          m_sC5);
	SetIniStr   ("C6",	          m_sC6);
	SetIniStr   ("C7",	          m_sC7);
	SetIniStr   ("C8",	          m_sC8);
 	SetIniStr   ("PlotEquation",  m_sFuncEquation);
 	SetIniStr   ("RedEquation",   m_sREquation);
 	SetIniStr   ("GreenEquation", m_sGEquation);
 	SetIniStr   ("BlueEquation",  m_sBEquation);

	// common parameters
	SetIniDouble("ZRotate",       m_zrotate);
	SetIniDouble("XYTilt",        m_xytilt);
	SetIniDouble("PlotScale",     m_scale);
	SetIniLong  ("LineColor",     m_line_color);
	SetIniLong  ("BackColor",     m_back_color);
	SetIniLong  ("ShowColor",     m_is_color);
	SetIniLong  ("RemoveHidden",  m_remove_hidden);
 	SetIniLong  ("ShowAxis",      m_show_axis);

	// specific parameters
	SetIniDouble("XMin",          m_xmin);
	SetIniDouble("XMax",          m_xmax);
	SetIniDouble("YMin",          m_ymin);
	SetIniDouble("YMax",          m_ymax);
	SetIniDouble("ZMin",          m_zmin);
	SetIniDouble("ZMax",          m_zmax);
 	SetIniLong   ("DrawLines",    m_draw_lines);
	SetIniLong   ("nZPlanes",     m_nzplanes);
	SetIniLong   ("UseSpherical", m_use_sphere_range);
	SetIniDouble ("SphereXCenter",m_sphere_xcenter);
	SetIniDouble ("SphereYCenter",m_sphere_ycenter);
	SetIniDouble ("SphereZCenter",m_sphere_zcenter);
	SetIniDouble ("SphereRadius", m_sphere_radius);
	SetIniLong   ("ClipMode",     m_clip_mode);
	SetIniLong   ("InvertClip",   m_is_clip_inverted);

	return(0);
}

/*	<><><><><><><><><><><><><>  Plot3DF.cpp  <><><><><><><><><><><><><><> */















